/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "tcp/tcpipv6segment.h"

#include <logfunc.hpp>

#include "boost_assert_handler.h"

using namespace std;
using boost::asio::ip::address;
using boost::date_time::time_resolution_traits_adapted64_impl;
using boost::posix_time::ptime;
using boost::posix_time::microsec_clock;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// TcpIpv6Segment
//-----------------------------------------------------------------------------

/**
 * @brief Default constructor.
 */
TcpIpv6Segment::TcpIpv6Segment() :
    IpHeader(),
    TcpHdr()
{
}

/**
 * @brief Parameterized constructor.
 *
 * @param tcp_header The TCP header.
 */
TcpIpv6Segment::TcpIpv6Segment( const TcpHeader &tcp_header ) :
    IpHeader(),
    TcpHdr( tcp_header )
{
}

/**
 * @brief Destructor.
 */
TcpIpv6Segment::~TcpIpv6Segment()
{
}

/**
 * @brief Obtain the IP header.
 *
 * @return The IP header object.
 */
Ipv6Header TcpIpv6Segment::get_ip_header() const
{
    return IpHeader;
}

/**
 * @brief Obtain the TCP header.
 *
 * @return The TCP header object.
 */
TcpHeader TcpIpv6Segment::get_tcp_header() const
{
    return TcpHdr;
}

/**
 * @brief Convenience method to check if this segment, matching the arguments,
 * is a RST reply.
 *
 * @param source_address The source address.
 *
 * @return @c true if this packet is a RST reply, or @c false otherwise.
 */
bool TcpIpv6Segment::match_rst_reply(
        const address &source_address
) const
{
    BOOST_ASSERT( source_address.is_v6() );

    Ipv6Header ipv6_header = get_ip_header();
    TcpHeader tcp_header = get_tcp_header();

    bool match_reset = tcp_header.get_reset();
    bool match_address = ipv6_header.get_source_address() == source_address;

    return ( match_reset && match_address );
}

/**
 * @brief Prints the RST reply.
 *
 * @param time_segment_sent The time when this segment was sent.
 *
 * @return void
 */
void TcpIpv6Segment::print_rst_reply(
        const ptime &time_segment_sent
) const
{
    Ipv6Header ipv6_header = get_ip_header();
    TcpHeader tcp_header = get_tcp_header();

    uint32_t sequence_number = tcp_header.get_sequence_number();
    int hops = ipv6_header.get_hop_limit();
    ptime now = microsec_clock::universal_time();
    time_resolution_traits_adapted64_impl::int_type elapsed_time =
            (now - time_segment_sent).total_milliseconds();

    GlobalLogger.info() << "RST from " << ipv6_header.get_source_address()
        << ": tcp_seq=" << sequence_number
        << ", hops=" << hops
        << " time=" << elapsed_time << " ms"
        << endl;
}

/**
 * @brief Convenience method to check if this segment, matching the arguments,
 * is a SYN/ACK reply.
 *
 * @param source_address The source address.
 *
 * @return @c true if this packet is a SYN/ACK reply, or @c false otherwise.
 */
bool TcpIpv6Segment::match_syn_ack_reply(
        const address &source_address
) const
{
    BOOST_ASSERT( source_address.is_v6() );

    Ipv6Header ipv6_header = get_ip_header();
    TcpHeader tcp_header = get_tcp_header();

    bool match_synchronize = tcp_header.get_synchronize();
    bool match_acknowledgment = tcp_header.get_acknowledgment();
    bool match_address = ipv6_header.get_source_address() == source_address;

    return ( match_synchronize && match_acknowledgment && match_address );
}

/**
 * @brief Prints the SYN/ACK reply.
 *
 * @param time_segment_sent The time when this segment was sent.
 *
 * @return void
 */
void TcpIpv6Segment::print_syn_ack_reply(
        const ptime &time_segment_sent
) const
{
    Ipv6Header ipv6_header = get_ip_header();
    TcpHeader tcp_header = get_tcp_header();

    uint32_t sequence_number = tcp_header.get_sequence_number();
    int hops = ipv6_header.get_hop_limit();
    ptime now = microsec_clock::universal_time();
    time_resolution_traits_adapted64_impl::int_type elapsed_time =
            (now - time_segment_sent).total_milliseconds();

    GlobalLogger.info() << "SYN/ACK from " << ipv6_header.get_source_address()
        << ": tcp_seq=" << sequence_number
        << ", hops=" << hops
        << " time=" << elapsed_time << " ms"
        << endl;
}

/**
 * @brief Read the TCP Segment from the @c istream.
 *
 * @param is The input stream.
 *
 * @return @c true if the read was successful, or @c false if an error occurred.
 */
bool TcpIpv6Segment::read( istream &is )
{
    is.clear();

    is >> *this;

    return !is.fail();
}

/**
 * @brief Write the TCP Segment to the @c ostream.
 *
 * @param os The output stream.
 *
 * @return @c true if the write was successful, or @c false if an error occurred.
 */
bool TcpIpv6Segment::write( ostream &os ) const
{
    os.clear();

    os << *this;

    return !os.fail();
}

istream& operator>>(
        istream &is,
        TcpIpv6Segment &segment
)
{
    is >> segment.IpHeader >> segment.TcpHdr;

    return is;
}

ostream& operator<<(
        ostream& os,
        const TcpIpv6Segment &segment
)
{
    os << segment.TcpHdr;

    return os;
}
